package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.logger.getLogger
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel


interface IWorker {
    /**
     * Adds a new task to the worker's internal tasks channel.
     */
    suspend fun addTask(task: Task)

    // TODO: check if we can use autoclosable interface here.
    /**
     * Function used to close the underlying resources.
     */
    fun close()
}


fun getCustomThreadpoolWorker(): Worker {
    val numCores = Runtime.getRuntime().availableProcessors()
    // use a fixed size thread pool context with numCores threads
    // TODO: Threadpool needs to be closed
    @OptIn(DelicateCoroutinesApi::class)
    val threadPoolContext = newFixedThreadPoolContext(numCores, "CustomThreadWorkerPool")
    val coroutineScope = CoroutineScope(threadPoolContext)
    return Worker("Custom-ThreadPool-Worker", coroutineScope)
}

/**
 * Abstract worker class used as a base for multithreaded worker.
 * It uses a fixed thread pool as a base for the coroutineScope.
 * The size of the thread pool is equal to the number of cores.
 */
class Worker(private val name: String,
             private val coroutineScope: CoroutineScope = CoroutineScope(Dispatchers.Default),
             numberOfWorkers: Int = Runtime.getRuntime().availableProcessors()
): IWorker {

    // TODO: this could be dangerous to give it unlimited memory
    private val tasks = Channel<Task>(Channel.UNLIMITED)


    private val logger = getLogger(javaClass)

    override suspend fun addTask(task: Task) {
        tasks.send(task)
    }

    init {
        logger.info("Starting $numberOfWorkers workers.")
        repeat(numberOfWorkers) {
            logger.info("[${Thread.currentThread().name}] launching coroutine in $name")
            launchWorker(it)
        }
    }

    /**
     * Wrapper function to execute tasks provided through the channel.
     * We use a dedicated number of coroutines to watch the tasks channel.
     * Each coroutine starts a new coroutine to execute the task. This is
     * done so that the outer coroutines are not blocked by the potentially
     * very long-running inner task performed by executeTask.
     */
    private fun launchWorker(id: Int) = coroutineScope.launch {
        for (task in tasks) {
            coroutineScope.launch {
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name received $task")
                // TODO: we need to check if we want to wrap this executeTask in a try-catch block
                task.execute()
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name finished $task")
            }
        }
    }

    override fun close() {
        coroutineScope.launch {
            tasks.close()
            for (job in coroutineContext[Job]!!.children) {
                job.join()
            }
            logger.info("Worker $name has been closed")
        }
    }
}
